
#include <mforms/mforms.h>
#include <algorithm>

#include <grt/grt_manager.h>

#include "grt_wizard_form.h"

using namespace grtui;
using namespace mforms;

#define WIZARD_WIDTH 800
#define WIZARD_HEIGHT 600


WizardForm::WizardForm(bec::GRTManager *mgr)
: _grtm(mgr), _toplevel(false), 
  _header_back(false), _header(false), _contents(false), _button_box(true)
{
  set_content(&_toplevel);

  _active_page= 0;

  set_size(WIZARD_WIDTH, WIZARD_HEIGHT);

  _top_title.set_style(BoldStyle);

  _header.set_padding(12);
  _header.set_spacing(4);
  _header.add(&_top_title, false, false);
  _header.add(&_top_subtitle, false, false);

  _header_back.set_back_color("#ffffff");
  _header_back.set_size(-1, 60);
  _header_back.add(&_header);

  _button_box.set_padding(12);
  _button_box.set_spacing(8);

  _button_box.add_end(&_next, false, false);
  _button_box.add_end(&_back, false, false);
  _button_box.add_end(&_cancel, false, false);
  _button_box.add(&_advanced, false, false);

  _cancel.set_text(_("_Cancel"));
  _cancel.set_size(100, -1);
  _cancel.signal_clicked().connect(sigc::mem_fun(this, &WizardForm::cancel_cb));

  _next.set_text(_("_Next >"));
  _next.set_size(100, -1);
  _next.signal_clicked().connect(sigc::mem_fun(this, &WizardForm::go_to_next));

  _back.set_text(_("< _Back"));
  _back.set_size(100, -1);
  _back.signal_clicked().connect(sigc::mem_fun(this, &WizardForm::go_to_back));

  _advanced.set_text(_(" _Advanced >> "));
  _advanced.signal_clicked().connect(sigc::mem_fun(this, &WizardForm::advanced_cb));

  _toplevel.add(&_header_back, false, true);
  _toplevel.add(&_contents, true, true);
  _toplevel.add(&_button_box, false, true);


  _advanced.show(false);
  
  _values= grt::DictRef(mgr->get_grt());
}


WizardForm::~WizardForm()
{
}


int WizardForm::default_width() const
{
  return WIZARD_WIDTH;
}


void WizardForm::add_page(WizardPage *page)
{
  _pages.push_back(page);
}


grt::DictRef WizardForm::show(grt::DictRef &values)
{
  switch_to_page(_pages.front(), true);

  show_modal(0, 0);
  
  return values;
}


void WizardForm::switch_to_page(WizardPage *page, bool advancing)
{
  if (_active_page)
  {
    if (page != _active_page)
    {
      _active_page->show(false);
      _active_page->leave(_values);
    }
  }

  if (_active_page != page)
  {
    if (advancing)
      page->load();

    if (_active_page)
      _contents.remove(_active_page);

    _contents.add(page);

    _active_page= page;

    _active_page->show(true);

    _active_page->enter(_values);

    update_buttons();
    
    if (advancing)
      _active_page->execute();
  }

  _top_title.set_text(_active_page->get_title());
  _top_subtitle.set_text(_active_page->get_subtitle());
  if (_active_page->get_advanced_panel())
    _advanced.show(true);
  else
    _advanced.show(false);

  update_buttons();
}


int WizardForm::get_active_page_number()
{
  return (int)(std::find(_pages.begin(), _pages.end(), _active_page) - _pages.begin());
}


void WizardForm::set_top_icon(const std::string &path)
{

}


void WizardForm::set_problem(const std::string &problem)
{
  _problem= problem;
  
  update_buttons();
}


void WizardForm::clear_problem()
{
  _problem.clear();
  update_buttons();
}


void WizardForm::update_buttons()
{
  if (_active_page)
  {
    std::string caption= _active_page->next_button_caption();
    if (caption.empty())
      caption= _("_Next >");
    _next.set_text(caption);

    _next.set_enabled(_problem.empty() && _active_page->allow_next());
    _back.set_enabled(!_turned_pages.empty() && _active_page->allow_back());
    _cancel.set_enabled(_active_page->allow_cancel());
  }
}


WizardPage *WizardForm::get_next_page(WizardPage *current)
{
  for (std::vector<WizardPage*>::const_iterator iter= _pages.begin();
       iter != _pages.end(); ++iter)
  {
    if (*iter == current)
    {
      ++iter;
      if (iter != _pages.end())
        return *iter;
      return 0;
    }
  }
  return 0;
}


void WizardForm::go_to_next()
{
  if (_active_page)
  {    
    if (_active_page->override_next_action() &&
        _active_page->execute_overriden_next_action())
      return;

    _next.set_enabled(false);
    _back.set_enabled(false);
    _cancel.set_enabled(false);

    _turned_pages.push_back(_active_page);
    
    if (_active_page->next_closes_wizard())
    {
      close();
      return;
    }

    WizardPage *next_page= get_next_page(_active_page);
    if (!next_page)
    {
      g_warning("Trying to turn past the last page of the wizard");
      cancel_cb();
      return;
    }
    switch_to_page(next_page, true);
  }
}


void WizardForm::go_to_back()
{
  if (!_turned_pages.empty())
  {
    WizardPage *back_page= _turned_pages.back();
    _turned_pages.pop_back();

    switch_to_page(back_page, false);
  }
}


void WizardForm::cancel_cb()
{
  close();
}


void WizardForm::advanced_cb()
{
  if (_active_page)
  {
    _active_page->show_advanced(!_active_page->advanced_shown());
    if (_active_page->advanced_shown())
      _advanced.set_text(" _Advanced << ");
    else
      _advanced.set_text(" _Advanced >> ");
  }
}


//--------------------------------------------------------------------------------

WizardPage::WizardPage(WizardForm *form, const std::string &id)
: Box(false), _form(form), _id(id)
{
  _advanced_shown= false;
  set_padding(12);
  set_spacing(12);
}


void WizardPage::show_advanced(bool flag)
{
  _advanced_shown= flag;
  get_advanced_panel()->show(flag);
}

bool WizardPage::advanced_shown() const
{
  return _advanced_shown;
}


bool WizardPage::override_next_action()
{
  return false;
}


bool WizardPage::execute_overriden_next_action()
{
  return false;
}


void WizardPage::enable_file_browsing(mforms::TextEntry *entry, mforms::Button *button, mforms::FileChooserType type)
{
  button->signal_clicked().
    connect(sigc::bind<mforms::TextEntry*, mforms::FileChooserType>(sigc::mem_fun(this, &WizardPage::browse_file_callback),
                                                                    entry, type));
}


void WizardPage::browse_file_callback(mforms::TextEntry *entry, mforms::FileChooserType type)
{
  FileChooser fsel(type);

  if (fsel.show_modal())
    entry->set_value(fsel.get_path());
}



void WizardPage::validate()
{
  if (_form->get_active_page() == this)
  {
    do_validate();
  }
  _form->update_buttons();
}



void WizardPage::enter(grt::DictRef values)
{
  _signal_enter.emit(values);
}


void WizardPage::leave(grt::DictRef values)
{
  _signal_leave.emit(values);
}

