
#include "preferences_form.h"





PreferencesForm::PreferencesForm(wb::WBContextUI *wbui, const std::string &model_id)
  : _top_box(false), _bottom_box(true), _button_box(true),

    _font_list(true)
{
  _wbui= wbui;
  _model_id= model_id;

  set_title(_("Workbench Preferences"));

  set_size(600, 520);
  set_content(&_top_box);
  _top_box.set_padding(0);
  _top_box.set_spacing(0);

  _top_box.add(&_tabview, true, true);
  
  _top_box.add(&_bottom_box, false);

  _bottom_box.add_end(&_button_box, false);
  _button_box.set_padding(12);
  _button_box.set_spacing(8);
  _button_box.set_homogeneous(true);

  _ok_button.signal_clicked().connect(sigc::mem_fun(this, &PreferencesForm::ok_clicked));
  _cancel_button.signal_clicked().connect(sigc::mem_fun(this, &PreferencesForm::cancel_clicked));

  _button_box.add_end(&_cancel_button, false, true);
  _cancel_button.set_text(_("Cancel"));

  _button_box.add_end(&_ok_button, false, true);
  _ok_button.set_text(_(" OK "));

  create_general_page();
  create_model_page();
  create_mysql_page();
  create_diagram_page();
  create_appearance_page();
  
  show_values();
}


void PreferencesForm::show_values()
{
  for (std::list<Option*>::const_iterator iter= _options.begin(); iter != _options.end(); ++iter)
    (*iter)->show_value();
  
  show_colors();
}


void PreferencesForm::update_values()
{
  _wbui->begin_wb_options_change();
  
  for (std::list<Option*>::const_iterator iter= _options.begin(); iter != _options.end(); ++iter)
    (*iter)->update_value();
  
  update_colors();
  
  _wbui->end_wb_options_change();
}


grt::DictRef PreferencesForm::get_options()
{
  if (_model_id.empty())
    return _wbui->get_wb()->get_wb_options();
  else
    return _wbui->get_model_options(_model_id);
}


static void show_entry_option(grt::DictRef options, const std::string &option_name, mforms::TextEntry *entry, bool numeric)
{
  if (numeric)
    entry->set_value(options.get(option_name).repr());
  else
    entry->set_value(options.get_string(option_name));
}


static void update_entry_option(grt::DictRef options, const std::string &option_name, mforms::TextEntry *entry, bool numeric)
{
  if (numeric)
    options.gset(option_name, atol(entry->get_string_value().c_str()));
  else
    options.gset(option_name, entry->get_string_value());
}


static void show_checkbox_option(grt::DictRef options, const std::string &option_name, mforms::CheckBox *checkbox)
{
  checkbox->set_active(options.get_int(option_name));
}


static void update_checkbox_option(grt::DictRef options, const std::string &option_name, mforms::CheckBox *checkbox)
{
  options.gset(option_name, checkbox->get_active());
}


static void show_selector_option(grt::DictRef options, const std::string &option_name, mforms::Selector *selector,
                                 const std::vector<std::string> &choices)
{
  selector->set_selected(std::find(choices.begin(), choices.end(), options.get_string(option_name)) - choices.begin());
}


static void update_selector_option(grt::DictRef options, const std::string &option_name, mforms::Selector *selector,
                                   const std::vector<std::string> &choices)
{
  if (selector->get_selected_index() < 0)
    options.gset(option_name, "ask");
  else
    options.gset(option_name, choices[selector->get_selected_index()]);
}


mforms::TextEntry *PreferencesForm::new_entry_option(const std::string &option_name, bool numeric)
{
  Option *option= new Option();
  mforms::TextEntry *entry= new mforms::TextEntry();
  grt::DictRef options(get_options());

  option->view= entry;
  option->show_value= sigc::bind(sigc::ptr_fun(&show_entry_option), options, option_name, entry, numeric);
  option->update_value= sigc::bind(sigc::ptr_fun(&update_entry_option), options, option_name, entry, numeric);
  _options.push_back(option);

  return entry;
}


mforms::CheckBox *PreferencesForm::new_checkbox_option(const std::string &option_name)
{
  Option *option= new Option();
  mforms::CheckBox *checkbox= new mforms::CheckBox();
  grt::DictRef options(get_options());

  option->view= checkbox;
  option->show_value= sigc::bind(sigc::ptr_fun(&show_checkbox_option), options, option_name, checkbox);
  option->update_value= sigc::bind(sigc::ptr_fun(&update_checkbox_option), options, option_name, checkbox);
  _options.push_back(option);

  return checkbox;
}


mforms::Selector *PreferencesForm::new_selector_option(const std::string &option_name)
{
  Option *option= new Option();
  mforms::Selector *selector= new mforms::Selector();
  grt::DictRef options(get_options());
  std::string choices_string;

  _wbui->get_wb_options_value(_model_id, "@"+option_name+"/Items", choices_string);
  std::vector<std::string> choices, parts= bec::split_string(choices_string, ",");

  for (std::vector<std::string>::const_iterator iter= parts.begin();
       iter != parts.end(); ++iter)
  {
    std::vector<std::string> tmp= bec::split_string(*iter, ":", 1);
    if (tmp.size() == 1)
    {
      selector->add_item(*iter);
      choices.push_back(*iter);
    }
    else
    {
      selector->add_item(tmp[0]);
      choices.push_back(tmp[1]);
    }
  }

  option->view= selector;
  option->show_value= sigc::bind(sigc::ptr_fun(&show_selector_option), options, option_name, selector, choices);
  option->update_value= sigc::bind(sigc::ptr_fun(&update_selector_option), options, option_name, selector, choices);
  _options.push_back(option);

  return selector;
}


static void destroy_option(PreferencesForm::Option *object)
{
  delete object;
}


static void destroy_view(mforms::View *object)
{
  delete object;
}


PreferencesForm::~PreferencesForm()
{
  std::for_each(_widgets.begin(), _widgets.end(), sigc::ptr_fun(destroy_view));
  std::for_each(_options.begin(), _options.end(), sigc::ptr_fun(destroy_option));
}


mforms::Label *PreferencesForm::new_label(const std::string &text, bool right_align)
{
  mforms::Label *label= add_view(new mforms::Label());
  label->set_text(text);
  if (right_align)
    label->set_text_align(mforms::MiddleRight);

  return label;
}

void PreferencesForm::ok_clicked()
{
  update_values();

  show(false);
  _handle_close();
}


void PreferencesForm::cancel_clicked()
{
  show(false);
  _handle_close();
}


void PreferencesForm::set_closed_handler(const sigc::slot<void> &slot)
{
  _handle_close= slot;
}


void PreferencesForm::create_general_page()
{
  mforms::Table *table= add_view(new mforms::Table());

  table->set_padding(18);
  table->set_row_spacing(8);
  table->set_column_spacing(4);
  table->set_row_count(2);
  table->set_column_count(2);
  _widgets.push_back(table);

  {
    mforms::Box *hbox= add_view(new mforms::Box(true));
    mforms::TextEntry *entry= new_entry_option("workbench:UndoEntries", true);
    mforms::Label *descr= add_view(new mforms::Label());
    
    hbox->set_spacing(4);

    //label->set_size(200, -1);
    entry->set_max_length(5);
    entry->set_size(50, -1);
    descr->set_text(_("Set to 0 for unlimited undo."));

    hbox->add(entry, false);
    hbox->add(descr, true, true);
    
    

    table->add(new_label(_("Undo History Size:"), true), 0, 1, 0, 1, mforms::HFillFlag);
    table->add(hbox, 1, 2, 0, 1, mforms::HFillFlag);
  }
  
  {
    mforms::CheckBox *check= new_checkbox_option("workbench.AutoReopenLastModel");
    check->set_text(_("Automatically Reopen Previous Model When Started"));
    table->add(check, 1, 2, 1, 2, mforms::HFillFlag);
  }
   
  _tabview.add_page(table, _("General"));
}


void PreferencesForm::create_model_page()
{
  mforms::Box *box= add_view(new mforms::Box(false));
  box->set_padding(12);
  box->set_spacing(8);
  _widgets.push_back(box);


  {
    mforms::Panel *frame= add_view(new mforms::Panel(true));
    frame->set_title(_("When Deleting Physical Model Figures in Diagram"));
    
    mforms::Box *rbox= add_view(new mforms::Box(true));
    
    rbox->set_padding(8);
    
    frame->add(rbox);

    rbox->add(new_label(""), false);
    rbox->add(new_selector_option("workbench.physical:DeleteObjectConfirmation"), true, false);

    box->add(frame, false);
  }

  {
    mforms::Panel *frame= add_view(new mforms::Panel(true));
    frame->set_title(_("Primary Key Defaults"));

    mforms::Box *hbox= add_view(new mforms::Box(true));

    hbox->set_padding(8);
    hbox->set_spacing(4);

    frame->add(hbox);

    mforms::TextEntry *entry;
    
    hbox->add(new_label(_("Column Name:"), true), false);
    entry= new_entry_option("PkColumnNameTemplate", false);
    entry->set_tooltip(_("Substitutions:\n"
                         "%table% - name of the table"));
    hbox->add(entry, true, true);

    hbox->add(new_label(_("Column Type:"), true), false);
    entry= new_entry_option("PkColumnType", false);
    entry->set_tooltip(_("Default type for use in newly added primary key columns.\nSpecify a column type name or a user defined type.\nFlags such as UNSIGNED are not accepted."));
    hbox->add(entry, true, true);

    box->add(frame, false);
  }

  {
    mforms::Panel *frame= add_view(new mforms::Panel(true));
    frame->set_title(_("Foreign Key/Relationship Defaults"));

    mforms::Table *table= add_view(new mforms::Table());

    table->set_padding(8);

    frame->add(table);
    
    table->set_row_spacing(4);
    table->set_column_spacing(4);    
    table->set_row_count(3);
    table->set_column_count(4);

    mforms::TextEntry *entry;

    table->add(new_label(_("FK Name:"), true), 0, 1, 0, 1, mforms::HFillFlag);
    entry= new_entry_option("FKNameTemplate", false);
    entry->set_tooltip(_("Substitutions:\n"
                         "%table%, %stable% - name of the source table\n"
                         "%dtable% - name of the destination table (where FK is added)\n"
                         "%column%, %scolumn% - name of the source column"));
    table->add(entry, 1, 2, 0, 1, mforms::HFillFlag|mforms::HExpandFlag);

    table->add(new_label(_("Column Name:"), true), 2, 3, 0, 1, mforms::HFillFlag);
    entry= new_entry_option("FKColumnNameTemplate", false);
    entry->set_tooltip(_("Substitutions:\n"
                         "%table%, %stable% - name of the source table\n"
                         "%dtable% - name of the destination table (where FK is added)\n"
                         "%column%, %scolumn% - name of the source column"));
    table->add(entry, 3, 4, 0, 1, mforms::HFillFlag|mforms::HExpandFlag);

    table->add(new_label(_("ON UPDATE:"), true), 0, 1, 1, 2, mforms::HFillFlag);
    table->add(new_selector_option("db.ForeignKey:updateRule"), 1, 2, 1, 2, mforms::HFillFlag|mforms::HExpandFlag);

    table->add(new_label(_("ON DELETE:"), true), 2, 3, 1, 2, mforms::HFillFlag);
    table->add(new_selector_option("db.ForeignKey:deleteRule"), 3, 4, 1, 2, mforms::HFillFlag|mforms::HExpandFlag);


    table->add(new_label(_("Associative Table Name:"), true), 0, 1, 2, 3, mforms::HFillFlag);
    entry= new_entry_option("AuxTableTemplate", false);
    entry->set_tooltip(_("Substitutions:\n"
                         "%stable% - name of the source table\n"
                         "%dtable% - name of the destination table"));
    table->add(entry, 1, 2, 2, 3, mforms::HFillFlag|mforms::HExpandFlag);

    table->add(new_label(_("for n:m relationships")), 2, 4, 2, 3, mforms::HFillFlag);

    box->add(frame, false);
  }

  _tabview.add_page(box, _("Model"));
}


void PreferencesForm::create_mysql_page()
{
  mforms::Box *box= add_view(new mforms::Box(false));
  box->set_padding(12);
  box->set_spacing(8);
  _widgets.push_back(box);


  {
    mforms::Panel *frame= add_view(new mforms::Panel(true));
    frame->set_title(_("Table Defaults"));

    mforms::Box *tbox= add_view(new mforms::Box(true));

    tbox->set_padding(8);

    frame->add(tbox);

    tbox->add(new_label(_("Default Storage Engine:"), true), false, false);
    tbox->add(new_selector_option("db.mysql.Table:tableEngine"), true, true);
    
    box->add(frame, false);
  }

  _tabview.add_page(box, _("MySQL"));
}



void PreferencesForm::create_diagram_page()
{
  mforms::Box *box= add_view(new mforms::Box(false));
  box->set_padding(12);
  box->set_spacing(8);
  _widgets.push_back(box);


  {
    mforms::Panel *frame= add_view(new mforms::Panel(true));
    frame->set_title(_("All Objects"));

    mforms::Box *vbox= add_view(new mforms::Box(false));

    vbox->set_padding(8);
    vbox->set_spacing(4);

    frame->add(vbox);
    
    mforms::CheckBox *check;

    check= new_checkbox_option("workbench.physical.ObjectFigure:Expanded");
    check->set_text(_("Expand New Objects"));
    check->set_tooltip(_("Set the initial state of newly created objects to expanded (or collapsed)"));
    vbox->add(check, false);

    check= new_checkbox_option("SynchronizeObjectColors");
    check->set_text(_("Propagate Object Color Changes to All Diagrams"));
    check->set_tooltip(_("If an object figure's color is changed, all figures in all diagrams that represent the same object are also updated"));
    vbox->add(check, false);
    
    box->add(frame, false);
  }

  {
    mforms::Panel *frame= add_view(new mforms::Panel(true));
    frame->set_title(_("Tables"));

    mforms::Box *vbox= add_view(new mforms::Box(false));

    vbox->set_padding(8);
    vbox->set_spacing(4);

    frame->add(vbox);
    
    mforms::CheckBox *check;

    check= new_checkbox_option("workbench.physical.TableFigure:ShowColumnTypes");
    check->set_text(_("Show Column Types"));
    check->set_tooltip(_("Show the column types along their names in table figures"));
    vbox->add(check, false);

    box->add(frame, false);
  }


  {
    mforms::Panel *frame= add_view(new mforms::Panel(true));
    frame->set_title(_("Routines"));

    mforms::Box *hbox= add_view(new mforms::Box(true));

    hbox->set_padding(8);
    hbox->set_spacing(4);

    frame->add(hbox);
    
    mforms::TextEntry *entry;

    hbox->add(new_label(_("Trim Routine Names Longer Than")), false);
    
    entry= new_entry_option("workbench.physical.RoutineGroupFigure:MaxRoutineNameLength", true);
    entry->set_size(60, -1);
    entry->set_max_length(3);
    hbox->add(entry, false);

    hbox->add(new_label(_("characters")), false);

    box->add(frame, false);
  }


  {
    mforms::Panel *frame= add_view(new mforms::Panel(true));
    frame->set_title(_("Relationships/Connections"));

    mforms::Box *vbox= add_view(new mforms::Box(false));

    vbox->set_padding(8);
    vbox->set_spacing(4);

    frame->add(vbox);
    
    mforms::CheckBox *check;

    check= new_checkbox_option("workbench.physical.Diagram:DrawLineCrossings");
    check->set_text(_("Draw Line Crossings (slow in large diagrams)"));
    vbox->add(check, false);

    check= new_checkbox_option("workbench.physical.Connection:HideCaptions");
    check->set_text(_("Hide Captions"));
    vbox->add(check, false);

    check= new_checkbox_option("workbench.physical.Connection:CenterCaptions");
    check->set_text(_("Center Captions Over Line"));
    vbox->add(check, false);

    box->add(frame, false);
  }

  _tabview.add_page(box, _("Diagram"));
}



static void show_text_option(grt::DictRef options, const std::string &option_name, mforms::TextBox *text)
{
  text->set_value(options.get_string(option_name));
}


static void update_text_option(grt::DictRef options, const std::string &option_name, mforms::TextBox *text)
{
  options.gset(option_name, text->get_string_value());
}


void PreferencesForm::create_appearance_page()
{
  mforms::Box *box= add_view(new mforms::Box(false));
  box->set_padding(12);
  box->set_spacing(8);
  _widgets.push_back(box);

  
  {
    mforms::Panel *frame= add_view(new mforms::Panel(true));
    frame->set_title(_("Color Presets"));

    mforms::Table *table= add_view(new mforms::Table());

    table->set_padding(8);
    table->set_row_spacing(4);
    table->set_column_spacing(4);    
    table->set_row_count(2);
    table->set_column_count(2);

    frame->add(table);

    mforms::TextBox *text;
    
    table->add(new_label(_("Colors available for tables, views etc")), 0, 1, 0, 1, mforms::HFillFlag);
    text= add_view(new mforms::TextBox());
    table->add(text, 0, 1, 1, 2, mforms::HFillFlag|mforms::VFillFlag|mforms::VExpandFlag|mforms::HExpandFlag);
    
    Option *option= new Option();
    _options.push_back(option);
    option->view= text;
    option->show_value= sigc::bind(sigc::ptr_fun(show_text_option), get_options(), "workbench.model.ObjectFigure:ColorList", text);
    option->update_value= sigc::bind(sigc::ptr_fun(update_text_option), get_options(), "workbench.model.ObjectFigure:ColorList", text);
    
    table->add(new_label(_("Colors available for layers, notes etc")), 1, 2, 0, 1, mforms::HFillFlag);
    text= add_view(new mforms::TextBox());
    table->add(text, 1, 2, 1, 2, mforms::HFillFlag|mforms::VFillFlag|mforms::VExpandFlag|mforms::HExpandFlag);

    option= new Option();
    _options.push_back(option);
    option->view= text;
    option->show_value= sigc::bind(sigc::ptr_fun(show_text_option), get_options(), "workbench.model.Figure:ColorList", text);
    option->update_value= sigc::bind(sigc::ptr_fun(update_text_option), get_options(), "workbench.model.Figure:ColorList", text);


    box->add(frame, false);
  }

  {
    mforms::Panel *frame= add_view(new mforms::Panel(true));
    frame->set_title(_("Fonts"));

    mforms::Table *table= add_view(new mforms::Table());

    table->set_padding(8);
    table->set_row_spacing(4);
    table->set_column_spacing(4);    
    table->set_row_count(2);
    table->set_column_count(2);

    frame->add(table);

    _font_list.add_column(mforms::StringColumnType, _("Location"), false);
    _font_list.add_column(mforms::StringColumnType, _("Font"), true);
    _font_list.end_columns();

    table->add(&_font_list, 0, 1, 0, 1, mforms::HFillFlag|mforms::VFillFlag|mforms::VExpandFlag|mforms::HExpandFlag);

    box->add(frame, true, true);
  }


  _tabview.add_page(box, _("Appearance"));
}


static std::string separate_camel_word(const std::string &word)
{
  std::string result;

  for (std::string::const_iterator c= word.begin(); c != word.end(); ++c)
  {
    if (!result.empty() && *c >= 'A' && *c <= 'Z')
      result.append(" ");
    result.append(1, *c);
  }
  
  return result;
}


void PreferencesForm::show_colors()
{
  std::vector<std::string> options= _wbui->get_wb_options_keys("");

  _font_options.clear();
  _font_list.clear_rows();
  
  for (std::vector<std::string>::const_iterator iter= options.begin();
       iter != options.end(); ++iter)
  {
    if (bec::has_suffix(*iter, "Font") && bec::has_prefix(*iter, "workbench."))
    {
      std::string::size_type pos= iter->find(':');
      
      if (pos != std::string::npos)
      {
        try
        {
          std::string part= iter->substr(pos + 1);
          std::string figure= bec::split_string(iter->substr(0, pos), ".")[2];
          std::string caption;

          part= part.substr(0, part.length() - 4);

          caption= separate_camel_word(figure) + " " + part;
          
          int row= _font_list.add_row();
          std::string value;
          _wbui->get_wb_options_value("", *iter, value);
          _font_list.set(row, 0, caption);
          _font_list.set(row, 1, value);
          
          _font_options.push_back(*iter);
        }
        catch (...)
        {
        }
      }
    }
  }
}


void PreferencesForm::update_colors()
{
  for (int c= _font_list.count(), i= 0; i < c; i++)
  {
    std::string value= _font_list.get_string(i, 1);

    _wbui->set_wb_options_value("", _font_options[i], value);
  }
}


