//
//  WBMenuManager.mm
//  MySQLWorkbench
//
//  Created by Alfredo Kojima on 3/Oct/08.
//  Copyright 2008 Sun Microsystems Inc. All rights reserved.
//

#import "WBMenuManager.h"

#include "incremental_list_updater.h"


// In MacOS X the following items are in the app menu so they are not managed by the menu manager:
// - About
// - Preferences...
// - Quit
// They 

@interface WBMenuItem : NSMenuItem
{
  NSString *_command;
}

- (id)initWithItem:(const bec::MenuItem&)item action:(SEL)action;
- (NSString*)command;
- (void)update:(const bec::MenuItem&)item;
@end

@implementation WBMenuItem

- (id)initWithItem:(const bec::MenuItem&)item action:(SEL)action
{
  self= [super initWithTitle:[NSString stringWithUTF8String:item.caption.c_str()]
                      action:action
               keyEquivalent:@""];
  if (self != nil)
  {
    _command= [[NSString stringWithUTF8String:item.name.c_str()] retain];
    
    if (!item.shortcut.empty())
    {
      std::vector<std::string> parts(bec::split_string(item.shortcut, "+", 0));
      std::string key;
      NSUInteger mask= 0;
      
      if (parts.size() > 0)
      {
        key= parts.back();
        parts.pop_back();
        
        if (key.length() == 1)
        {
          key[0]= tolower(key[0]);
        }
        else
        {
          if (key == "Delete")
            key= NSDeleteCharacter;
          else if (key == "BackSpace" || key == "Backspace")
            key= NSBackspaceCharacter;
          else
            NSLog(@"Unknown character '%s'", key.c_str());
        }
        
        [self setKeyEquivalent:[NSString stringWithUTF8String:key.c_str()]];
      }
  
      if (parts.size() > 0)
      {
        for (std::vector<std::string>::const_iterator iter= parts.begin();
             iter != parts.end(); ++iter)
        {
          if (*iter == "Command")
            mask|= NSCommandKeyMask;
          else if (*iter == "Alt" || *iter == "Alternate" || *iter == "Option")
            mask|= NSAlternateKeyMask;
          else if (*iter == "Control")
            mask|= NSControlKeyMask;
          else if (*iter == "Shift")
            mask|= NSShiftKeyMask;
        }
        if (mask != 0)
          [self setKeyEquivalentModifierMask:mask];
      }
    }
    
    [self update:item];
  }
  return self;
}


- (void)update:(const bec::MenuItem&)item
{
  if (strcmp([[self title] UTF8String], item.caption.c_str())!=0)
    [self setTitle:[NSString stringWithUTF8String:item.caption.c_str()]];
  
  [self setEnabled:item.enabled];
  if (item.type == bec::MenuCheck || item.type == bec::MenuRadio)
    [self setState:item.checked ? NSOnState : NSOffState];  
}


- (NSString*)command
{
  return _command;
}

- (void)dealloc
{
  [_command release];
  [super dealloc];
}

@end



@implementation WBMenuManager


class IncrementalMenuRebuilder 
: public bec::IncrementalListUpdater<NSUInteger,wb::MenuItemList::const_iterator>
{
  WBMenuManager *_manager;
  NSMenu *_menu;
  NSArray *_menuItems;
  const wb::MenuItemList &_items;
  
public:
  IncrementalMenuRebuilder(WBMenuManager *manager, NSMenu *menu, const wb::MenuItemList &items)
  : _manager(manager), _menu(menu), _items(items)
  {
    _menuItems= [[_menu itemArray] retain];
  }
  
  virtual ~IncrementalMenuRebuilder()
  {
    [_menuItems release];
  }
  
  virtual dest_iterator get_dest_iterator()
  {
    return 0;
  }
  
  virtual source_iterator get_source_iterator()
  {
    return _items.begin();
  }
  
  virtual dest_iterator increment_dest(dest_iterator &iter)
  {
    return ++iter;
  }
  
  virtual source_iterator increment_source(source_iterator &iter)
  {
    return ++iter;
  }
  
  virtual bool has_more_dest(dest_iterator iter)
  {
    return iter < [_menuItems count];
  }
  
  virtual bool has_more_source(source_iterator iter)
  {
    return iter != _items.end();
  }
  
  virtual bool items_match(dest_iterator diter, source_iterator siter)
  {
    NSMenuItem *item= [_menuItems objectAtIndex:diter];
    
    return (strcmp(siter->caption.c_str(), [[item title] UTF8String])==0);
  }
  
  virtual void remove(dest_iterator iter)
  {
    NSMenuItem *item= [_menuItems objectAtIndex:iter];
    
    if ([item isKindOfClass:[WBMenuItem class]])
      [_menu removeItem:item];
  }
  
  virtual void insert(dest_iterator dest_pos, source_iterator source_item)
  {
    NSMenuItem *newItem;
    
    if (source_item->type == bec::MenuSeparator)
    {
      newItem= [NSMenuItem separatorItem];
    }
    else
    {
      newItem= [[[WBMenuItem alloc] initWithItem:*source_item action:@selector(activateItem:)] autorelease];
      [newItem setTarget:_manager];
    }
    
    if (source_item->type == bec::MenuCascade)
    {
      NSMenu *submenu= [[NSMenu alloc] initWithTitle:[newItem title]];
      wb::MenuItemList subItems= _manager->_wbui->get_menu_items(source_item->name);
      [newItem setSubmenu:submenu];
      
      [submenu setAutoenablesItems:NO];
      
      if (!subItems.empty())
      {
        IncrementalMenuRebuilder rebuilder(_manager, submenu, subItems);
        
        rebuilder.execute();
      }
    }
    
    [_menu insertItem:newItem atIndex:dest_pos];
  }
  
  virtual void update(dest_iterator dest_item, source_iterator source_item)
  {
    NSMenuItem *item= [_menuItems objectAtIndex:dest_item];
    
    if ([item isKindOfClass:[WBMenuItem class]])
      [(WBMenuItem*)item update:*source_item];
    
    if ([item submenu])
    {
      wb::MenuItemList subItems= _manager->_wbui->get_menu_items(source_item->name);
      
      IncrementalMenuRebuilder rebuilder(_manager, [item submenu], subItems);
      
      rebuilder.execute();
    }
  }
};




- (id)initWithMenuBar:(NSMenu*)menu
            WBContext:(wb::WBContextUI*)wbui
{
  if ((self= [super init]))
  {
    _menuBar= menu;
    _wbui= wbui;
  }
  return self;
}


- (void)activateItem:(WBMenuItem*)item
{
  _wbui->activate_command([[item command] UTF8String]);
}



- (void)differentialMenuRebuild:(NSMenu*)menu
                          items:(const wb::MenuItemList&)items
{
  IncrementalMenuRebuilder rebuilder(self, menu, items);
  
  rebuilder.execute();
}


- (void)rebuildMainMenu
{
  // temporarily remove 1st item from menubar
  NSMenuItem *firstItem= [[_menuBar itemAtIndex:0] retain];
  [_menuBar removeItemAtIndex:0];
  [self differentialMenuRebuild:_menuBar items:_wbui->get_main_menus()];
  
  [_menuBar insertItem:firstItem atIndex:0];
  [firstItem release];
}


- (void)refreshMenu:(NSMenu*)menu
{
  /*
  for (NSMenuItem* item in [menu itemArray])
  {
    
  }*/
  
  
  
}

@end
