//
//  MDCanvasView.mm
//  mdcanvas
//
//  Created by Alfredo Kojima on 07/Mar/6.
//  Copyright 2007 MySQL AB. All rights reserved.
//

#import "MDCanvasView.h"
#import "MDCanvasScrollView.h"
#include <sys/time.h>

static void *create_context(CGLPixelFormatObj, void*, CGLContextObj*);
static void destroy_context(void*);
static void copy_context(void*,void*,unsigned long mask);
static void set_drawable(void*, void*);
static void *get_drawable(void*);


static glitz_cgl_callbacks_t glitz_callbacks= {
  create_context,
  destroy_context,
  copy_context,
  set_drawable,
  get_drawable
};


class CanvasViewDelegate
{
  MDCanvasView *_view;
public:
  CanvasViewDelegate(MDCanvasView *view) 
  : _view(view)
  {
    mdc::CanvasView *cv= [view mdCanvasView];
    
    cv->signal_repaint().connect(sigc::mem_fun(this, &CanvasViewDelegate::needs_repaint));
    cv->signal_viewport_changed().connect(sigc::mem_fun(this, &CanvasViewDelegate::viewport_changed));
  }
  
  void needs_repaint()
  {
    [_view setNeedsDisplay:YES];
  }

  void needs_repaint2(int x, int y, int width, int height)
  {
    [_view setNeedsDisplayInRect:NSMakeRect(x, y, width, height)];
  }
  
  
  void viewport_changed()
  {
    [(MDCanvasScrollView*)[_view enclosingScrollView] reflectContentRect];
  }
};


@implementation MDCanvasView

+ (void)initialize
{ 
  [super initialize];

  mdc::MacOSXCanvasView::init_glitz_backend(glitz_callbacks);
}


- (id)initWithFrame:(NSRect)frame 
{
  self = [super initWithFrame:frame];
  if (self)
  {
    _canvas= new mdc::MacOSXCanvasView(0, self, (int)NSWidth([self frame]), (int)NSHeight([self frame]));
  
    _delegate= new CanvasViewDelegate(self);
    
    [[NSNotificationCenter defaultCenter] addObserver:self
                                             selector:@selector(frameDidChange:)
                                                 name:NSViewFrameDidChangeNotification
                                               object:self];
    _initialized= NO;
  }
  return self;
}


- (void)dealloc
{
  delete _canvas;
  [super dealloc];
}


- (BOOL)isFlipped
{
  return YES;
}


- (void)viewDidMoveToWindow
{
  [[self window] makeFirstResponder:self];
  [[self window] setAcceptsMouseMovedEvents:YES];
}


- (void)drawRect:(NSRect)rect 
{
  if (!_initialized)
  {
    // can't init earlier or we get "invalid drawable"
    _initialized= YES;
    _canvas->initialize();
  }

  timeval tv, tv2;
  
  gettimeofday(&tv, 0);
  
  _canvas->repaint((int)NSMinX(rect), (int)NSMinY(rect), (int)NSWidth(rect), (int)NSHeight(rect));
  
  gettimeofday(&tv2, 0);
  
  printf("rendertime= %.4f (%.1ffps)\n", (tv2.tv_sec - tv.tv_sec) + (tv2.tv_usec - tv.tv_usec)/1000000.0,
         1.0/((tv2.tv_sec - tv.tv_sec) + (tv2.tv_usec - tv.tv_usec)/1000000.0));
}


- (void)frameDidChange:(NSNotification*)notif
{ 
  NSRect frame= [self frame];
  if (_canvas && _initialized)
  {
    _canvas->update_view_size((int)NSWidth(frame), (int)NSHeight(frame));
    _canvas->repaint((int)NSMinX(frame), (int)NSMinY(frame), (int)NSWidth(frame), (int)NSHeight(frame));
    [_nscontext update];
  }
}


- (void)setOpenGLContext:(NSOpenGLContext*)context
{
  _nscontext= context;
}


- (mdc::CanvasView*)mdCanvasView
{
  return _canvas;
}


- (void)setOrigin:(NSPoint)point
{
  _canvas->set_offset(mdc::Point(point.x, point.y));
}



static mdc::EventState getModifiers(NSEvent *event)
{
  int mods= [event modifierFlags];
  mdc::EventState res= mdc::SNone;
  
  if (mods & NSShiftKeyMask)
    res= res | mdc::SShiftMask;
  if (mods & NSCommandKeyMask)
    res= res | mdc::SCommandMask;
  if (mods & NSControlKeyMask)
    res= res | mdc::SControlMask;
  if (mods & NSAlternateKeyMask)
    res= res | mdc::SAltMask;
  
  return res;
}


- (void)mouseDown:(NSEvent*)event
{
  NSPoint pos= [self convertPoint:[event locationInWindow] fromView:nil];
  
  _canvas->handle_mouse_button((mdc::MouseButton)[event buttonNumber], true, (int)pos.x, (int)pos.y, 
                               getModifiers(event));
}


- (void)mouseUp:(NSEvent*)event
{
  NSPoint pos= [self convertPoint:[event locationInWindow] fromView:nil];
  
  _canvas->handle_mouse_button((mdc::MouseButton)[event buttonNumber], false, (int)pos.x, (int)pos.y, 
                               getModifiers(event));
}



- (void)mouseMoved:(NSEvent*)event
{
  [self mouseDragged:event];
}


- (void)mouseDragged:(NSEvent*)event
{
  NSPoint pos= [self convertPoint:[event locationInWindow] fromView:nil];

  _canvas->handle_mouse_move((int)pos.x, (int)pos.y, getModifiers(event));
}


- (NSSize)totalSize
{
  mdc::Size size= _canvas->get_total_view_size();
  
  return NSMakeSize(size.width, size.height);
}


- (NSRect)visibleRect
{
  mdc::Rect rect= _canvas->get_viewport();
  
  return NSMakeRect(rect.xmin(), rect.ymin(), rect.size.width, rect.size.height);
}


- (MDCanvasScrollView*)enclosingScrollView
{
  return (MDCanvasScrollView*)[self superview];
}

@end



//==============================================================================


static NSOpenGLPixelFormat *nspixelFormatFromCGL(CGLPixelFormatObj fmt)
{
  struct ConvTable {
    CGLPixelFormatAttribute attr;
    NSOpenGLPixelFormatAttribute nsattr;
    BOOL extra;
  } cattr[]= {
  { kCGLPFADoubleBuffer, NSOpenGLPFADoubleBuffer, 0 },
  { kCGLPFAAlphaSize, NSOpenGLPFAAlphaSize, 1 },
  { kCGLPFAColorSize, NSOpenGLPFAColorSize, 1 },
  { kCGLPFAStencilSize, NSOpenGLPFAStencilSize, 1 },
  { kCGLPFASampleBuffers, NSOpenGLPFASampleBuffers, 1 },
  { kCGLPFASamples, NSOpenGLPFASamples, 1 },
  { (CGLPixelFormatAttribute)0 }
  };
  NSOpenGLPixelFormatAttribute attribs[2*(sizeof(cattr)/sizeof(ConvTable))+1];
  int i, j;
  long value;
  
  for (j= i= 0; cattr[i].attr != 0; i++)
  {
    if (CGLDescribePixelFormat(fmt, 0, cattr[i].attr, &value) == kCGLNoError)
    {
      attribs[j++]= cattr[i].nsattr;
      if (cattr[i].extra)
      {
        attribs[j++]= (NSOpenGLPixelFormatAttribute)value;
      }
    }
  }
  attribs[j]= (NSOpenGLPixelFormatAttribute)0;
  
  return [[NSOpenGLPixelFormat alloc] initWithAttributes:attribs];
}


static void *create_context(CGLPixelFormatObj fmt, void *share, CGLContextObj *retcontext)
{
  NSOpenGLPixelFormat *format= nspixelFormatFromCGL(fmt);
  
  NSOpenGLContext *nscontext= [[NSOpenGLContext alloc] initWithFormat:format
                                                         shareContext:(NSOpenGLContext*)share];
  
  *retcontext= (CGLContextObj)[nscontext CGLContextObj];
  
  return nscontext;
}


static void destroy_context(void *context)
{
  [(NSOpenGLContext*)context release];
}


static void copy_context(void *srccontext, void *dstcontext, unsigned long mask)
{
  [(NSOpenGLContext*)dstcontext copyAttributesFromContext:(NSOpenGLContext*)srccontext
                                                 withMask: mask];
  
}


static void set_drawable(void *context, void *view)
{
  MDCanvasView *canvas= (MDCanvasView*)view;
  [canvas setOpenGLContext: (NSOpenGLContext*)context];
  [(NSOpenGLContext*)context setView:canvas];
}


static void *get_drawable(void *context)
{
  return [(NSOpenGLContext*)context view];
}

