using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Data;
using System.Drawing;
using System.Text;
using System.Windows.Forms;
using WeifenLuo.WinFormsUI.Docking;
using MySQL.Grt;
using MySQL.Utilities;
using MySQL.Grt.Db;

namespace MySQL.GUI.Workbench.Plugins
{
  public partial class DbMysqlRoutineEditor : ObjectEditorPlugin
  {
    #region Member Variables

    private MySQLRoutineEditorBE routineEditorBE { get { return editorBE as MySQLRoutineEditorBE; } } 

    private DbObjectEditorPages dbObjectEditorPages;

    protected Timer sqlEditorTimer;

    #endregion

    #region Constructors

    public DbMysqlRoutineEditor()
    {
      InitializeComponent();
    }

    public DbMysqlRoutineEditor(GrtManager GrtManager, GrtValue GrtList)
      : base(new MySQLRoutineEditorBE(GrtManager, GrtList))
    {
      InitializeComponent();

      // Init BE
      routineEditorBE.set_sql_parser_log_cb(new DBObjectEditorBE.VoidStringListDelegate(sqlEditorParseLogCallback));

      // Init timer
      sqlEditorTimer = new Timer();
      sqlEditorTimer.Interval = 2000;
      sqlEditorTimer.Tick += new EventHandler(sqlEditorTimerHandler);

      InitializingControls = true;
      try
      {
        InitFormData();

        CallRefreshFormData();
      }
      finally
      {
        InitializingControls = false;
      }
      routineEditorBE.add_refresh_ui_handler(RefreshFormDataInvoke);
    }

    #endregion

    #region IWorkbenchDocument implementation

    public override UIForm BackendClass
    {
      get { return routineEditorBE; }
    }

    #endregion

    #region ObjectEditorPlugin Overrides

    public override bool ChangeGrtList(GrtManager GrtManager, GrtValue GrtList)
    {
      // Make sure the changes are stored
      DbMysqlRoutineEditor_FormClosing(null, null);

      InitializingControls = true;

      try
      {
        grtManager = GrtManager;
        grtList = GrtList;

        if (editorBE != null)
          editorBE.disable_auto_refresh();
        editorBE = new MySQLRoutineEditorBE(GrtManager, GrtList);

        InitFormData();

        RefreshFormData();
      }
      finally
      {
        InitializingControls = false;
      }

      routineEditorBE.add_refresh_ui_handler(RefreshFormDataInvoke);

      Invalidate();

      return true;
    }

    #endregion

    #region Form implementation

    private void DbMysqlRoutineEditor_Load(object sender, EventArgs e)
    {
      ddlEditor.GrtManager = grtManager;
      ddlEditor.SqlEditorBE = routineEditorBE.get_sql_editor();
      ddlEditor.Initialize();

      ActiveControl = ddlEditor;
    }

    private void DbMysqlRoutineEditor_FormClosing(object sender, FormClosingEventArgs e)
    {
      sqlEditorTimerHandler(null, null);
    }

    protected void InitFormData()
    {
      dbObjectEditorPages = new DbObjectEditorPages(grtManager, routineEditorBE);
    }

    protected override void RefreshFormData()
    {
      nameTextBox.Text = routineEditorBE.get_name();
      TabText = nameTextBox.Text;

      String ddl = routineEditorBE.get_sql_definition_header();
      String routineSQL = routineEditorBE.get_sql();
      int caretPos = ddlEditor.Caret.Position;
      if (0 < routineSQL.Length)
      {
        ddl += routineSQL + "//\n";

        syntaxCheckLabel.Visible = true;
        if (routineEditorBE.has_syntax_error())
        {
          syntaxCheckLabel.Text = "DDL Syntax Error";
          syntaxCheckLabel.Image = Properties.Resources.parse_indicator_red;
          syntaxCheckLabel.Cursor = Cursors.Hand;
        }
        else
        {
          syntaxCheckLabel.Text = "DDL Syntax OK";
          syntaxCheckLabel.Image = Properties.Resources.parse_indicator_green;
          syntaxCheckLabel.Cursor = Cursors.Default;
        }
      }
      else
      {
        //! todo: this should be moved to BE, split into two strings so cursor can be set
        ddl += String.Format("CREATE PROCEDURE `{0}`.`{1}` ()\nBEGIN\n  ",
          routineEditorBE.get_schema_name(), routineEditorBE.get_name());
        caretPos = ddl.Length;
        ddl += "\nEND//\n";
      }
      ddlEditor.Text = ddl;
      ddlEditor.Caret.Position = caretPos;
      ddlEditor.Selection.Start = caretPos;
      ddlEditor.Selection.End = caretPos;
    }

    private void sqlEditorTimerHandler(Object myObject, EventArgs myEventArgs)
    {
      if (sqlEditorTimer.Enabled)
      {
        sqlEditorTimer.Stop();
        routineEditorBE.set_sql(ddlEditor.Text);
      }
    }

    private void ddlEditor_DocumentChange(object sender, ScintillaNet.NativeScintillaEventArgs e)
    {
      if (!InitializingControls && !insideRefreshFormData)
      {
        sqlEditorTimer.Stop();
        sqlEditorTimer.Start();
      }
    }

    private void ddlEditor_TextModified(object sender, ScintillaNet.TextModifiedEventArgs e)
    {
      ddlEditor_DocumentChange(sender, null);
    }

    private void sqlEditorParseLogCallback(List<String> strlist)
    {
      // if there are more than 2 lines, the 2nd line contains the first error
      if (strlist.Count >= 3)
        MySQLApplicationContext.MainFormStatusBarText = strlist[1];
      else
        MySQLApplicationContext.MainFormStatusBarText = "";

      // name field could change...
      if (routineEditorBE.get_name() != nameTextBox.Text)
        nameTextBox.Text = routineEditorBE.get_name();

    }

    #endregion

    private void syntaxCheckLabel_Click(object sender, EventArgs e)
    {
      if (routineEditorBE.has_syntax_error())
        grtManager.get_grt().make_output_visible();
    }

    private void undoItem_Click(object sender, EventArgs e)
    {
      ddlEditor.UndoRedo.Undo();
    }

    private void cutItem_Click(object sender, EventArgs e)
    {
      ddlEditor.Clipboard.Cut();
    }

    private void copyItem_Click(object sender, EventArgs e)
    {
      ddlEditor.Clipboard.Copy();
    }

    private void pasteItem_Click(object sender, EventArgs e)
    {
      ddlEditor.Clipboard.Paste();
    }

    private void selectAllItem_Click(object sender, EventArgs e)
    {
      ddlEditor.Selection.SelectAll();
    }

    private void redoItem_Click(object sender, EventArgs e)
    {
      ddlEditor.UndoRedo.Redo();
    }

    private void editorContextMenu_Opening(object sender, CancelEventArgs e)
    {
      // Enable/Disable menu items depending on the edit content.
      undoItem.Enabled = ddlEditor.UndoRedo.CanUndo;
      redoItem.Enabled = ddlEditor.UndoRedo.CanRedo;
      cutItem.Enabled = ddlEditor.Selection.Length > 0;
      copyItem.Enabled = ddlEditor.Selection.Length > 0;
      pasteItem.Enabled = ddlEditor.Clipboard.CanPaste;
      selectAllItem.Enabled = ddlEditor.TextLength > 0;
    }

  }
}