#ifndef _TEST_HELPERS_H_
#define _TEST_HELPERS_H_

#ifndef _WIN32
#include <vector>
#include <list>
#include <stdexcept>
#include <fstream>
#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <assert.h>
#ifndef EXCLUDE_MYSQL_SUPPORT
#include <my_global.h>
#include <mysql.h>
#endif
#ifdef IMAGE_TESTS
#include <cairo.h>
#endif
#endif

#ifdef bool
#undef bool
#endif

// needed for class Auto_release
#if 0
#include "myx_public_interface.h"
#endif

/*
#define ENUM_SIZE_EQUALS_INT_SIZE(enm) (sizeof(enm) == sizeof(int))

#if !ENUM_SIZE_EQUALS_INT_SIZE(library_error)
#error sizeof(enum) < sizeof(int)
#endif
*/

/*
  Global_test_parameters keeps global test environment runtime parameters
  such as test serer connection information
*/

class Global_test_parameters
{
public:
  // result output mode
  enum Output_mode { DEBUGGER, CONSOLE };

  // execution mode
  enum Run_mode { ALL, LIST, SELECTED_ONLY, SELECTED_SKIP };
  
private:
  // connection parameters
  const char *host_name;
  short port;
  const char *user_name;
  const char *password;
  std::vector <std::string> pathes;
  std::vector <std::string> groups;

  // reporting parameters
  bool rep_passed;
  Output_mode out_mode;
  bool wait_on_finish;
  bool color_output;

  Run_mode run_mode;

  void print_usage(const char *exe_name);
public:
  Global_test_parameters(int argc, char *argv[]);
  ~Global_test_parameters();

  std::string get_full_filename(const std::string& name);
  const char *get_host_name() const;
  short get_port() const;
  const char *get_user_name() const;
  const char *get_password() const;
  bool report_passed() const;
  Output_mode get_output_mode() const;
  bool wait() { return wait_on_finish; };
  Run_mode get_run_mode() const { return run_mode; };
  std::vector<std::string> get_groups() const { return groups; };
};

#ifndef EXCLUDE_MYSQL_SUPPORT
class Resultset;

/*
  Test_connection encapsulates one active test server connection.
  This class serves as factory for result set instances.
*/

class Test_connection
{
  MYSQL *mysql;

public:
   Test_connection(const char *host, short port, const char *user, const char *pass);
  ~Test_connection();

  // Try to avoid using get_mysql() if that's possible. if you lack a function that uses server connection
  // consider adding it to the Test_connection class.
  MYSQL* get_mysql(void) { return mysql; };

  void query(const std::string& query);

  // In multi_query the query string can contain several delimited statements and DELIMITER statements.
  void multi_query(std::istream& is);
  void multi_query(const std::string& q);
  Resultset* new_resultset(const std::string& query);
};

class Resultset
{
  std::vector<char**> rows;
  MYSQL_RES *res;
  int column_count;
  MYSQL_FIELD* fields;
  bool is_active;

  friend bool operator == (const Resultset &rs1, const Resultset &rs2);

protected:
  bool fetch_resultset();

public:
  Resultset(MYSQL_RES *resultset);
  ~Resultset();

  bool compare_to(const Resultset &resultset);
  char *escape(const char *str);
  int get_column_count(void) {return column_count; };
  MYSQL_FIELD* get_column_fields(void) {return fields; };
  bool load(const std::string &file);
  bool save(const std::string &file);
  char *unescape(const char *str);
};

#endif // EXCLUDE_MYSQL_SUPPORT

// Every TEST_FUNCTION is a member of a Test_object_base-derived class instance. The class could contain any shared info.
//  Note: template class is fake, it's just to get different template specializations.
namespace tut
{

template<class> class Test_object_base
{
};

}

// Test_group class encapsulates testgroup-wide information. For example it keeps a server connection
// which can be shared among the tests in group.
template<class module>
class Test_group : public ::tut::test_group< ::tut::Test_object_base<module> >
{
#ifndef EXCLUDE_MYSQL_SUPPORT
  Test_connection *c_;
#endif
  Global_test_parameters **params_;

  void inc_if_disabled();

public:
  Test_group(Global_test_parameters **params, const char* name, const char *desc)
    : tut::test_group<tut::Test_object_base<module> >(name),
#ifndef EXCLUDE_MYSQL_SUPPORT
      c_(NULL),
#endif
      params_(params)
  {
    inc_if_disabled();
  }

  virtual ~Test_group()
  {
#ifndef EXCLUDE_MYSQL_SUPPORT
    delete c_;
#endif
  }
#ifndef EXCLUDE_MYSQL_SUPPORT
  Test_connection *get_connection()
  {
    if (c_ == NULL)
    {
      c_= new Test_connection(
        (*params_)->get_host_name(),
        (*params_)->get_port(),
        (*params_)->get_user_name(),
        (*params_)->get_password());
    }
    return c_;
  }
#endif

  std::string get_full_filename(const std::string& name)
  {
    return (*params_)->get_full_filename(name);
  }

  bool compare_files(const std::string& created, const std::string& reference)
  {
    std::fstream file1(created.c_str(), std::ios::in | std::ios::binary);
    std::string name2 = (*params_)->get_full_filename(reference);
    std::fstream file2(name2.c_str(), std::ios::in | std::ios::binary);
    int line= 0;

    bool result= true;
    if (file1 && file2)
    {
      unsigned char *data1= new char[200000];
      unsigned char *data2= new char[200000];
      int current_offset= 0;

      while (!file1.eof() && !file2.eof())
      {
        file1.read(data1, 200000);
        int size1= file1.gcount();
        file2.read(data2, 200000);
        int size2= file2.gcount();
        if (size1 != size2)
        {
          int count= size1 < size2 ? size1 : size2;
          int i= 0;
          while ((count > 0) && (data1[i] == data2[i]))
          {
            --count;
            ++i;
          };

          int point= current_offset + i;
          std::cout << "Data differs at position: " << point << '\n';
          result= false;
          break;
        }
        else
        {

          int i= 0;
          while ((size1 > 0) && (data1[i] == data2[i]))
          {
            --size1;
            ++i;
          };
          if (size1 > 0)
          {
            int point= current_offset + i;
            std::cout << "Data differs at position: " << point << '\n';
            result= false;
            break;
          };
        };

        current_offset += size1;
      };

      delete[] data1;
      delete[] data2;
    }
    else
      result= false;
      
    if (!file1.eof() || !file2.eof())
      result= false;

    file1.close();
    file2.close();
    
    return result;
  }

#ifndef DISABLE_IMAGE_TESTS
  bool compare_images(const std::string& created, const std::string& reference)
  {
    cairo_surface_t *cimage= cairo_image_surface_create_from_png(created.c_str());
    cairo_surface_t *rimage= cairo_image_surface_create_from_png(reference.c_str());
    bool result= true;

    if (cairo_surface_status(cimage) != CAIRO_STATUS_SUCCESS)
    {
      std::cout << "Could not load PNG file " << created << '\n';
      result= false;
    }

    if (cairo_surface_status(rimage) != CAIRO_STATUS_SUCCESS)
    {
      std::cout << "Could not load reference PNG file " << reference << '\n';
      result= false;
    }
    
    if (result)
    {
      unsigned int cstride, ch;
      unsigned int rstride, rh;
      
      cstride= cairo_image_surface_get_stride(cimage);
      ch= cairo_image_surface_get_height(cimage);

      rstride= cairo_image_surface_get_stride(rimage);
      rh= cairo_image_surface_get_height(rimage);
      
      if (cstride != rstride || ch != rh)
      {
        std::cout << created << " and " << reference << " images have different sizes\n";
        result= false;
      }
      else
      {
        unsigned char *cdata, *rdata;
        
        cdata= cairo_image_surface_get_data(cimage);
        rdata= cairo_image_surface_get_data(rimage);

        if (memcmp(cdata, rdata, ch * cstride)!=0)
        {
          std::cout << created << " and " << reference << " images have different contents\n";
          result= false;
        }
      }
    }

    if (cimage)
      cairo_surface_destroy(cimage);
    if (rimage)
      cairo_surface_destroy(rimage);

    return result;
  }
#endif
};

template<class module> void Test_group<module>::inc_if_disabled()
{
  tut::test_group<tut::Test_object_base<module> >:: set_disabled(false);
}

/*
#ifdef _WIN32
template<> void Test_group<void>::inc_if_disabled()
{
  inc_disabled_groups_count();
  set_disabled(true);
}
#else
*/
template<> void Test_group<void>::inc_if_disabled();
//#endif

// @class ScopeGuardBase
// @brief Abstraction class for ScopeGuard
struct ScopeGuardBase
{
  virtual ~ScopeGuardBase() {}
};

//////////////////////////////////////////////////////////////////////
// @class ScopeGuard
//////////////////////////////////////////////////////////////////////
// @brief Template class helps auto-release resources
//
// Pass pointer to resource object and pointer to function which will 
// release that object.
//
// @example
// int* p = (int*)malloc(sizeof(int));
// ScopeGuard<int, void (*)(void*)> _auto_releaser(p, &free); 
// // now p will be released after scope leave 
//////////////////////////////////////////////////////////////////////
template<class T, class A>
struct ScopeGuard : ScopeGuardBase
{
public:
  explicit ScopeGuard(T* _v, A _a) :
    v(_v), a(_a) {
  }

  ~ScopeGuard() {
    if (v)
      (*a)(v);
  }

  void reset() {
    v= NULL;
  }

private:
  T *v;
  A a;
  ScopeGuard();
  ScopeGuard(ScopeGuard&);
  ScopeGuard& operator =(ScopeGuard&);
};



// Auto_release can be used to automatically free dynamically allocated objects. The usage pattern is the following:
//
// void f()
// {
//   Auto_release ar;
//   char *c= ar.add_g_free((char *)g_malloc(1000));
// }
//
// now c will be automatically freed (g_free-d in this case) when ar goes out of scope.
//
// At the moment only a limited set of datatypes is supported, so feel free to extend the class as needed.

class Auto_release
{
  std::list<char *> pointers;
  std::list<char *> cpp_pointers;
  std::list<ScopeGuardBase*> guards;
#if 0
  std::list<MYX_DBM_SERVER_VERSION *> versions;
  std::list<MYX_DBM_TABLE_DATA *> tabledatas;
  std::list<MYX_DBM_DATATYPES *> datatypes;
  std::list<MYX_SCHEMA_STORED_PROCEDURES *> schema_sps;
  std::list<MYX_RESULTSET *> resultsets;
  std::list<MYX_TABLE_EXPORTER_INFO *> expinfos;
  std::list<MYX_ENGINES *> engines;
#endif
#ifndef EXCLUDE_MYSQL_SUPPORT
  typedef std::pair<Test_connection *, std::string> Connection_sql_pair;

  std::list<Connection_sql_pair> sql_list;
#endif
public:

  char *add_g_free(char *p) { pointers.push_back(p); return p; }
  char *add_delete(char *p) { cpp_pointers.push_back(p); return p; }

  // add_guart - parametrized auto-deallocator
  // e.g. as equivalent to: char *c= ar.add_g_free((char *)g_malloc(1000));
  // may use: char *c= ar.add_guard( (char *)g_malloc(1000), g_free ); 
  template<class T, class A>  
  T *add_guard(T* p, A action) { guards.push_back( new ScopeGuard<T, A>(p, action)); return p; }
#if 0
  MYX_DBM_SERVER_VERSION *add(MYX_DBM_SERVER_VERSION *v) { versions.push_back(v); return v; }
  MYX_DBM_TABLE_DATA *add(MYX_DBM_TABLE_DATA *d) { tabledatas.push_back(d); return d; }
  MYX_DBM_DATATYPES *add(MYX_DBM_DATATYPES *t) { datatypes.push_back(t); return t; }
  MYX_SCHEMA_STORED_PROCEDURES *add(MYX_SCHEMA_STORED_PROCEDURES *sp) { schema_sps.push_back(sp); return sp; }
  MYX_RESULTSET *add(MYX_RESULTSET *rs) { resultsets.push_back(rs); return rs; }
  MYX_TABLE_EXPORTER_INFO *add(MYX_TABLE_EXPORTER_INFO *info) { expinfos.push_back(info); return info; }
  MYX_ENGINES *add(MYX_ENGINES *data) { engines.push_back(data); return data; }
#endif
#ifndef EXCLUDE_MYSQL_SUPPORT
  void add_sql(Test_connection *conn, const char *filename) { sql_list.push_back(Connection_sql_pair(conn, std::string(filename))); }
#endif

  ~Auto_release() { clear(); }

  void clear()
  {
    for(std::list<char *>::iterator it= pointers.begin(); it != pointers.end(); it++)
      g_free(*it);
    for(std::list<char *>::iterator it= cpp_pointers.begin(); it != cpp_pointers.end(); it++)
      delete *it;
    for(std::list<ScopeGuardBase *>::iterator it= guards.begin(); it != guards.end(); ++it)
      delete *it;
#if 0
    // TODO: check for NULL before passing to myx_free_*
    for(std::list<MYX_DBM_SERVER_VERSION *>::iterator it= versions.begin(); it != versions.end(); it++)
      myx_dbm_free_server_version(*it);
    for(std::list<MYX_DBM_TABLE_DATA *>::iterator it= tabledatas.begin(); it != tabledatas.end(); it++)
      myx_dbm_free_table_data(*it);
    for(std::list<MYX_DBM_DATATYPES *>::iterator it= datatypes.begin(); it != datatypes.end(); it++)
      if (*it) myx_free_datatype(*it);
    for(std::list<MYX_SCHEMA_STORED_PROCEDURES *>::iterator it= schema_sps.begin(); it != schema_sps.end(); it++)
      myx_free_schema_sps(*it);
    for(std::list<MYX_RESULTSET *>::iterator it= resultsets.begin(); it != resultsets.end(); it++)
      myx_query_free_resultset(*it);
    for(std::list<MYX_TABLE_EXPORTER_INFO *>::iterator it= expinfos.begin(); it != expinfos.end(); it++)
      myx_free_table_exporter_info(*it);
    for(std::list<MYX_ENGINES *>::iterator it= engines.begin(); it != engines.end(); it++)
      if (*it) myx_free_engines(*it);
#endif
#ifndef EXCLUDE_MYSQL_SUPPORT
    for(std::list<Connection_sql_pair>::iterator it= sql_list.begin(); it != sql_list.end(); it++)
    {
      Connection_sql_pair pair= *it;
      pair.first->multi_query(pair.second);
    }
#endif
  }
};

#endif // _TEST_HELPERS_H_
